/******************************************************************************
NirMalam filter plugin for Avisynth by V.C.Mohan

  In the specified frame either automatically selects a suitable window or uses given window.
  omputes the max diff. 
  
  window size, location  controls the amount of noise reduction and blurring.
  If reduction of sharpness can be tolerated,larger values for ry, gu, by reduces noise more.

  The window if user specified, should be located in an area on frame where the color 
  and intensity variation is minimal, so that noise statistics can be determined correctly.
  
  Color format input may also have some effect. RGB and YUV inputs may come out differently.
  
 
  Author V.C.Mohan

   16 Dec, 2006, 16 Mar 2009 5 Oct 2015

   Copyright (C) <2006, 2015>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  

// use x64 platform
#define AVISYNTHPLUS
// use win32 platform
#define AVISYNTH_VERSION_2_6

#ifdef AVISYNTHPLUS
#include "avisynthPlus.h"
#include <string.h>
const AVS_Linkage *AVS_linkage = 0;	// for dynamic linkage
#else
#ifdef AVISYNTH_VERSION_2_6
#include "windows.h"
//#define AVS_LINKAGE_DLLIMPORT  // This is for static linkage, so not required
#include "avisynth_2_6.h"
const AVS_Linkage *AVS_linkage = 0;	// for dynamic linkage
#else
#include "windows.h
#include "avisynth.h"	// 2.5 version
#endif
#endif

#include <math.h>
class NirMalam : public GenericVideoFilter {
			// specified parameters
		
		int xgrid, ygrid;	// size of grid x and to be used. Must be submultiples of frame width and height

		bool  printDiff;		// displays variance of given window in an error message
		
		int		nf;			// frame number from which window statistics to be taken
		bool autoselect;	// auto window select
		int minwindow;		// size in pixels of minimum window for autoselect
							// coordinates if autoselect is false
		int lx, rx;			// window left x, right x,
		int wty, wby;		// window top y and bottom y coordinates

		int ry, gu, bv;		// process and soft control



			// global variables

		
		int ty, by;			// window top y and bottom y coordinates  will be different fro ty and by for RGB
							// from which global values are to be computed
		bool gotGlobal;		// control to ensure that this is done only once on first entry

		int wx, hy;			// size of window for auto search;

		int difbuf[3];		// variance for  3 color planes 3 * 10 values

	//	int * relativeOffsets;
		

  public:
							//Definitions of functions
    NirMalam(PClip _child,int xgrid_, int ygrid_,			
							bool printDiff_,
							int nf_,
							bool autoselect_,int minwindow_,
							int lx_, int rx_, int wty_, int wby_,
							int ry_, int gu_, int bv_,

							IScriptEnvironment* env) ;

	~NirMalam();		//destructor

	void getGlobals(int nf, IScriptEnvironment* env, int * dbuf);

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	
	int getTotalDiff ( const unsigned char *sp,const int spitch, 
							int kb, int leftx, int topy,int rightx,
							int boty);

	void NirMalamProcess(unsigned char * dp, const int dpitch,
							const unsigned char * sp, const int spitch,
							const int gvarsq, const int ht, const int wd,
							const int y_grid, const int x_grid, const int bytes );


	int getMinDiffWindow( const unsigned char *sp, int spitch,
							int ht, int wd,int hy, int wx,
							int y_grid, int x_grid, int bytes);

	int getDiffToUse( const unsigned char *sp, int spitch, int* hist,
							int hy, int wx,int y_grid, int x_grid,
							int bytes, int percent);

	void ProcessBorder( unsigned char * dp, int dpitch,
							const unsigned char * sp, int spitch,int gvarsq,
							int ht, int wd, int y_grid, int x_grid, int bytes);

	void ProcessGrid(unsigned char *dp, const int dpitch,
							const unsigned char *sp, const int spitch,const int bytes,
							const int x, const int y, const int gridx, const int gridy,
							int varsq);

	#include "Planar_2_5_or_2_6.cpp"

};	
/*************************************************
 * The following is the implementation 
 * of the defined functions.
 *************************************************/
//Here is the acutal constructor code used

NirMalam::NirMalam(PClip _child,
						int xgrid_, int ygrid_,			
						bool printDiff_,
						 int nf_,
						bool autoselect_,int minwindow_,
						int lx_, int rx_, int wty_, int wby_,
						int ry_, int gu_, int bv_,

						IScriptEnvironment* env) :

	GenericVideoFilter(_child),
						xgrid(xgrid_), ygrid(ygrid_),					
						printDiff(printDiff_),
						
						nf(nf_), autoselect(autoselect_),
						minwindow(minwindow_),
						lx(lx_),rx(rx_), wty(wty_), wby(wby_),
						ry(ry_), gu(gu_), bv(bv_)
	
{	
	
		
	gotGlobal=false;

	if ( ! autoselect)
	{
		if (vi.IsRGB() )
		{
			ty = vi.height - 1 - wby;
			by = vi.height - 1 - wty;
		}

		else
		{
			ty = wty;
			by = wby;
		}

		wx = rx - lx + 1;
		hy = by - ty + 1;

	}

	else
	{		
					// auto search window size. Make area atleast minwindow
	
		int wx1 = (vi.width/ 60) & 0xfffffffe;
		wx = (vi.width/ 20) & 0xfffffffe;
		int hy1 = (vi.height/60) & 0xfffffffe;
		hy = (vi.height/20) & 0xfffffffe;

		if(wx < 8) wx = 8;
		if(hy < 8) hy = 8;

		while ( wx * hy < minwindow)
		{
					// window is small. increase width and height 
			wx += wx1;
			hy += hy1;

			if(wx > vi.width - 2 * xgrid)
			{

				wx -= wx1;
				wx1 = 0;
			}


			if( hy > vi.height - 2 * ygrid)
			{
				hy -= hy1;
				hy1 = 0;
			}

			if( wx < 16 && hy < 16)
				break;
		}
	}

	getGlobals(nf, env, difbuf);

//	relativeOffsets = new int [ xgrid * ygrid];


}

	//..................................................................................

NirMalam::~NirMalam()
{
//	delete [] relativeOffsets;
}

//-----------------------------------------------------------------------------------------

int NirMalam::getTotalDiff ( const unsigned char *sp,const int spitch, 
				int kb, int leftx, int topy,int rightx,
				int boty)
{
				// Calculates  mean  and total of differance of the specified rectangular area of frame. 
				// The total diff value is returned .

				// calculate mean
	int pixels = (rightx - leftx + 1) * (boty - topy + 1);

	int sum = pixels/2;	// to get to nearest integer

	for(int h = topy, hspitch = topy * spitch ; h <= boty; h ++, hspitch += spitch)

		for(int w = leftx, kbw = kb * leftx; w <= rightx; w ++, kbw += kb)

			sum += sp[hspitch + kbw];

	int lmean = sum /pixels;

				// get  differance 
	sum = 0;

	for(int h = topy, hspitch = topy * spitch ; h <= boty; h ++, hspitch += spitch)

		for(int w = leftx, kbw = kb * leftx; w <= rightx; w ++, kbw += kb)

				sum += abs( sp[hspitch + kbw] - lmean );

	return sum;	// returning total not divided by pixels


}

//-----------------------------------------------------------------------------------------------

int NirMalam::getMinDiffWindow( const unsigned char *sp, int spitch,
								int ht, int wd,int hy, int wx,
								int y_grid, int x_grid, int bytes)
{

						// finds a window which has minimum variance
	int minDif =  256 * wx * hy ;	// a high value not probable

						//we have a border of grid/2 so that we can process all of window later	
	for( int h = y_grid, hspitch = y_grid * spitch ; h < ht - hy - y_grid; h += hy, hspitch += spitch)
	{

		for( int w = x_grid, bytesw = bytes * x_grid; w < wd - wx - x_grid; w += wx, bytesw += bytes)

		{
						// calculate variance
			int wDif = getTotalDiff ( sp + hspitch + bytesw, spitch, bytes, 0, 0, wx -1, hy -1 );

			if (wDif < minDif)
			{
				minDif = wDif;
				lx = w;
				rx = w + wx - 1;

				ty = h;
				by = h + hy - 1;
			}

			if( minDif == 0)
						// a blank frame was given probably
				return 0;

		}
	}

	return minDif;	
}

//--------------------------------------------------------------------------------------------------

int NirMalam::getDiffToUse( const unsigned char *sp, int spitch, int* hist,
								int hy, int wx,int y_grid, int x_grid, int bytes, int percent)
{
	int count = 0, sum;

	int xy_grid = x_grid * y_grid, bytesx2 = x_grid/2 * bytes, spitchy2 = y_grid/2 * spitch;
						// initialize difference histogram buffer
	for(int i = 0; i < 256; i ++)
		hist[i] = 0;

	for( int h = 0, hspitch = 0; h < hy - y_grid; h ++, hspitch += spitch)

		for( int w = 0, bytesw = 0; w < wx - x_grid; w ++, bytesw += bytes)
		{
			sum = 0;

			for( int hh = h, hhspitch = hspitch; hh < h + y_grid; hh ++, hhspitch += spitch)

				for( int ww = w, bytesww = bytesw; ww < w + x_grid; ww ++, bytesww += bytes)

					sum += sp[ hhspitch + bytesww];

						// difference between mean and center value of grid

				int diff = abs(sum / (xy_grid) - sp[ hspitch + spitchy2 + bytesw + bytesx2] );

						// add to histogram

				hist[diff] ++ ;

				count ++;
		}

						// now get the nearest %age value desired 
						// hist[0] is for zero diff so not reqd	
		count -= hist[0];
		sum = hist[1];
		int i = 1;
		
		while ( sum < (count * percent)/100 )
		
			sum += hist[++ i];

		return (i);
}


//-------------------------------------------------------------------------------------------------

void NirMalam::NirMalamProcess(unsigned char * dp, const int dpitch,
							const unsigned char * sp, const int spitch,
							const int gdif, const int ht, const int wd,
							const int y_grid, const int x_grid, const int bytes )
{
					// main processing function
	int mean, sum, grid = x_grid * y_grid;
	
	int grid2 = grid/2;

	int bytesx2 = x_grid/2 * bytes, spitchy2 = y_grid/2 * spitch, dpitchy2 = y_grid/2 * dpitch;

	for(int h = 0, hspitch = 0, hdpitch = 0; h < ht - y_grid ; h ++, hspitch += spitch, hdpitch += dpitch)
	{

		sum = grid2;	// initialize for this row. To get nearest integer
		

		for ( int hh = h, hhspitch = hspitch; hh < h + y_grid; hh++, hhspitch += spitch)
		
			for( int ww = 0, bytesww = 0; ww < x_grid; ww ++, bytesww += bytes)

						// get mean for left most(grid) for this row
				sum += sp[hhspitch + bytesww];
				
		mean = sum /grid;

						//	int localDiff = abs(sp [(h + y_grid/2) * spitch + bytes *(x_grid/2)] - mean);

		if(abs(sp [(hspitch + spitchy2 + bytesx2)] - mean) > gdif)
						// local is high. So may be an edge. Ensure output is very near to this value
			dp [hdpitch + dpitchy2 + bytesx2]

				= sp [(hspitch + spitchy2 + bytesx2)]; 
							

		else

			dp [hdpitch + dpitchy2 + bytesx2] = mean;

						// process from second grid onwards to right of this row

		for(int w = 1, bytesw = bytes; w < wd - x_grid + 1; w ++, bytesw += bytes)
		{
				
			for ( int hh = h, hhspitch = hspitch; hh < h + y_grid; hh++, hhspitch += spitch)
			{
						// subtract left column			
				sum -= sp[hhspitch + bytesw - bytes];

						// add right column
				sum += sp[hhspitch + bytesw + bytesx2 - bytes];
			}
		

			mean = sum/grid;

			
			if(abs(sp [hspitch + spitchy2 + bytesw + bytesx2] - mean) > gdif)
						// local is high. So may be an edge. Ensure output is very near to this value
				dp [hdpitch + dpitchy2 + bytesw + bytesx2]

					= sp [hspitch + spitchy2 + bytesw + bytesx2]; 
						
			else
				dp [hdpitch + dpitchy2 + bytesw + bytesx2] = mean;
		}

	}
}



//-----------------------------------------------------------------------------------------------------------------
void NirMalam::ProcessGrid(unsigned char *dp, const int dpitch,
						   const unsigned char *sp, const int spitch,const int bytes,
						   const int x, const int y, const int gridx, const int gridy,
							int diff)
{
						// calculates mean and local dif in a grid
	int div = gridx * gridy;	// number of pixels in grid
	int sum = div >> 1;			// for nearest integer calculation
				
				// get mean
	for ( int ww = -gridx/2; ww <= gridx/2  ; ww ++)

		for ( int hh = - gridy/2; hh <= gridy / 2 ; hh ++)

							// get mean for left 
			sum += sp[abs(y + hh) * spitch + bytes * abs(x + ww) ];
				
	int mean = sum /div;
		
	if( abs(sp [y * spitch + bytes * x] - mean) > diff )
						 
						// local is high. So may be an edge. Ensure output is very near to this value
		dp [y * dpitch + bytes *  x ]

					= sp [y * spitch + bytes * x];  
					

	else

		dp [y * dpitch + bytes * x] = mean;
}

//---------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------
void NirMalam::ProcessBorder( unsigned char * dp, int dpitch, const unsigned char * sp, int spitch,int gvar,
						int ht, int wd, int y_grid, int x_grid, int bytes)
{


						// process left and right border xgrid/2 and ygrid/2 pixels

			
	for( int h = 0; h < ht - y_grid/2; h++)
	{
						// left border
		for( int w = 0; w < x_grid/2; w++)
		
			ProcessGrid(dp, dpitch, sp, spitch, bytes, w, h, x_grid, y_grid, gvar);

						//Right border

		for(int w = 0; w < x_grid/2; w++)
		
						// note - ve bytes

			ProcessGrid(dp + bytes * (wd - 1), dpitch, sp + bytes * (wd - 1), spitch,
										-bytes, w, h, x_grid, y_grid, gvar);


	}

						// process top and bottom borders

	for( int w = 0; w < wd - x_grid/2; w++)
	{
						// top border
		for( int h = 0; h < y_grid/2; h++)
		

			ProcessGrid(dp, dpitch, sp, spitch, bytes, w, h, x_grid, y_grid, gvar);

						// bottom border
				
		for(int h = 0; h < y_grid/2 + 1; h++)
		
						// note negative spitch, dpitch
			ProcessGrid(dp + (ht - 1) * dpitch, -dpitch, sp + (ht - 1) * spitch, -spitch,
						bytes, w, h, x_grid, y_grid, gvar);



	}

						//	we have four corners left out. just copy them

	for(int h = 0; h < y_grid/2; h++)
		for(int w = 0; w < x_grid/2; w++)
		{


			dp[ h * dpitch + w * bytes ] = sp[ h * spitch + w * bytes ];

			dp[ h * dpitch + (wd - 1 - w) * bytes ] = sp[ h * spitch + (wd - 1 - w) * bytes ];

			dp[ ( ht - 1 -h) * dpitch + w * bytes ] = sp[ ( ht - 1 -h) * spitch + w * bytes ];

			dp[ ( ht - 1 -h) * dpitch + (wd - 1 - w) * bytes ] = sp[ ( ht - 1 -h) * spitch + (wd - 1 - w) * bytes ];

			
		}
}

//---------------------------------------------------------------------------------------------------------------
void NirMalam::getGlobals(int nf, IScriptEnvironment* env, int * dbuf)
{
	
	PVideoFrame	Frame = child->GetFrame(nf, env);
	const unsigned char * fp = Frame->GetReadPtr();
	const	int fpitch = Frame->GetPitch();

	const unsigned char * fpU = Frame->GetReadPtr(PLANAR_U);
	const unsigned char * fpV = Frame->GetReadPtr(PLANAR_V);
	const	int fpitchUV = Frame->GetPitch(PLANAR_U);

	const	int bwd = vi.width;
    const	int bht = Frame->GetHeight();
	int kb = vi.BytesFromPixels(1);		

	int * histogram = new int[256];
	int color[] = {ry, gu, bv};

	if( vi.IsRGB() )
	{
		
		if(autoselect)
							// we use green color as it is nearest to Y 
			getMinDiffWindow( fp + 1, fpitch,bht, bwd, hy, wx,
									ygrid, xgrid, kb);

		for ( int p = 0; p < 3; p ++)
		{

			if(color[p] > 0)

				difbuf[p] = getDiffToUse(fp + p + ty * fpitch + kb * lx , fpitch,histogram,
								hy, wx, ygrid, xgrid, kb, color[p]);
			else

				difbuf[p] = 0;								

		}
	}

	else if(vi.IsYUY2() ) //  or YV12() )
	{
		int plus[] = {0, 1, 3};
		int wxx[] = {wx, wx / 2, wx / 2};
		int kbb[] = { kb, kb + kb, kb + kb};

		if(autoselect)

			getMinDiffWindow( fp, fpitch,
								bht, bwd, hy, wx,
								ygrid, xgrid, kb);

		for ( int p = 0; p < 3; p ++)
		{

			if(color[p] > 0)

				difbuf[p] = getDiffToUse(fp + plus[p] + ty * fpitch + kb * lx , fpitch,histogram,
								hy, wxx[p], ygrid, xgrid, kbb[p], color[p]);
			else

				difbuf[p] = 0;
		}
	}

	else	// Planar
	{
		int plane[] = {PLANAR_Y, PLANAR_U, PLANAR_V};
		int nplanes = IsY8() ? 1 : 3;

		for ( int p = 0; p < nplanes; p ++)
		{
			const unsigned char * fp = Frame->GetReadPtr();
			const	int fpitch = Frame->GetPitch();
			const	int bwd = Frame->GetRowSize(plane[p]);
			const	int bht = Frame->GetHeight(plane[p]);
			int subW = p == 0 ? 0 : GetPlaneWidthSubsampling(plane[p]);
			int subH = p == 0 ? 0 : GetPlaneHeightSubsampling(plane[p]);

			if ( p == 0)
			{
				if(autoselect)

					getMinDiffWindow( fp, fpitch,
										bht, bwd, hy, wx,
										ygrid, xgrid, kb);
			}

			if(color[p] > 0)

				difbuf[p] = getDiffToUse(fp + (ty >> subH) * fpitch + (lx >> subW), fpitch,histogram,
								hy >> subH, wx >> subW, ygrid, xgrid, 1, color[p]);
			else
				difbuf[p] = 0;

		}		

	}	// yuy or yv
				
	
		if(printDiff)
		{
			if( vi.IsRGB() )

				env->ThrowError("NirMalam: window lx = %d, rx = %d, ty = %d, by = %d, Differances R   %d   G   %d  B   %d",
								lx,rx,bht-1-by,bht-1-ty,
								difbuf[0], difbuf[1], difbuf[2]  );
							
			else 	//if( vi.Isyuy2 or planar

				env->ThrowError("NirMalam: window lx = %d, rx = %d, ty = %d, by = %d,Differances Y   %d  U   %d  V   %d  ",
								lx,rx,ty,by,
								difbuf[0], difbuf[1], difbuf[2]  );

		}

		delete []histogram;

		gotGlobal = true;

		
	}	
//---------------------------------------------------------------------------------------------------------------
PVideoFrame __stdcall NirMalam::GetFrame(int n, IScriptEnvironment* env)
{
	 

	
	PVideoFrame	Frame = child->GetFrame(n, env);
	const unsigned char * fp = Frame->GetReadPtr();
	const	int fpitch = Frame->GetPitch();	
	const	int bwd = vi.width;
    const	int bht = Frame->GetHeight();
	int kb = vi.BytesFromPixels(1);	

	

	if( difbuf[0] == 0 && difbuf[1] == 0  && difbuf[2] == 0 )
				// no noise  or blank frame.  Process is not required				
		return Frame;
	
				// Regular processing

	
	PVideoFrame work = env->NewVideoFrame(vi);
	unsigned char * wp = work->GetWritePtr();	
	int wpitch = work->GetPitch();

/*			// create offset values to use in grid to avoid multiplications
	for( int y = 0; y <  ygrid; y ++)

		for( int x = 0; x < xgrid; x ++)
		
			relativeOffsets[y * xgrid + x] = y * fpitch + kb * x;

*/

	if(vi.IsRGB() )
	{
		if( difbuf[0] == 0 || difbuf[1] == 0 || difbuf[2] == 0 )
			// copy frame as this is faster than copying individual color
			env->BitBlt(wp, wpitch, fp, fpitch, bwd * kb, bht);
		// actual processing:		
		for(int k = 0; k < 3; k++)
		{
			if( difbuf[k] > 0 )
			{
						// process to reduce variance (noise)
				NirMalamProcess(wp + k, wpitch, fp + k, fpitch,
						difbuf[k], bht, bwd, ygrid, xgrid, kb );
						
						// process border values as these were not processed. If more than iter 2 then need not as already has a copy
				
				ProcessBorder ( wp + k, wpitch, fp + k, fpitch, difbuf[k], bht, bwd, ygrid, xgrid, kb);	

			}			
	
		}	// for k = 0; k < 3		
		
	}	// if rgb
	
	else if(vi.IsYUY2())
	{
		if ( difbuf[0] == 0 || difbuf[1] == 0 || difbuf[2] == 0)
			// copy frame as bitblt is faster than any individual color copying if required
			env->BitBlt(wp, wpitch, fp, fpitch,bwd * kb, bht);
		int plus[] = {0, 1, 3};
		int kbb [] = {kb, kb + kb, kb + kb};
		int wd [] = {bwd, bwd / 2, bwd / 2};

		for( int p = 0; p < 3; p ++)
		{
			if( difbuf[p] > 0 )
			{
						// process to reduce variance (noise)
				NirMalamProcess(wp + plus[p] , wpitch, fp + plus[p] , fpitch,
						difbuf[p], bht, wd[p], ygrid, xgrid, kbb[p] );

						// process border values as these are not processed			
				ProcessBorder(wp + plus[p] , wpitch, fp + plus[p] , fpitch,
							difbuf[p], bht, wd[p], ygrid, xgrid, kbb[p] );

			}
		}			
	
	}	// YUY2

	else if(vi.IsPlanar())
	{
		int npl = IsY8() ? 1 : 3;
		int plane[] = {PLANAR_Y, PLANAR_U, PLANAR_V};
		
		for ( int p = 0; p < npl; p ++)
		{
			const unsigned char * fp = Frame->GetReadPtr(plane[p]);
			const	int fpitch = Frame->GetPitch(plane[p]);
			unsigned char * wp = work->GetWritePtr(plane[p]);
			const	int wpitch = work->GetPitch(plane[p]);
			int bwd = work->GetRowSize(plane[p]);
			int bht = work->GetHeight(plane[p]);

			if( difbuf[p] > 0  )
			{
							// process  to reduce variance (noise)
				NirMalamProcess(wp , wpitch, fp , fpitch,
								difbuf[p], bht, bwd, ygrid, xgrid, kb );

							// process border values as these are not processed
				ProcessBorder(wp , wpitch, fp , fpitch,
							difbuf[p], bht, bwd, ygrid, xgrid, kb );

			}	// 

			else
				// just copy plane				
				env->BitBlt(wp, wpitch, fp, fpitch, bwd, bht);
		}

	}	// planar	

	return work;		
}



/***************************************************************/
// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,

AVSValue __cdecl Create_NirMalam(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	const VideoInfo& vi= args[0].AsClip()->GetVideoInfo();

#ifdef AVISYNTH_VERSION_2_6
	if( !vi.IsPlanar() &&  !vi.IsRGB()  && !vi.IsYUY2() )
		env->ThrowError("NirMalam: This color format is not supported here");
#else
		if(!vi.IsRGB24() && !vi.IsRGB32()  && !vi.IsYV12() && !vi.IsYUY2())
			env->ThrowError("NirMalam: This color format is not supported by this old avisynth version");
#endif		
	if((args[1].AsInt(5)%2)==0 ||args[1].AsInt(5)<3 || args[1].AsInt(5)>11)
		env->ThrowError("NirMalam:xgrid should be odd positive number from 3 to 11");
	
	if((args[2].AsInt(5)%2)==0 ||args[2].AsInt(5)<3 || args[2].AsInt(5)>11)
		env->ThrowError("NirMalam:ygrid should be odd positive number from 3 to 11");
	
	if(args[4].AsInt(0) < 0 || args[4].AsInt(0) > vi.num_frames - 1)
		env->ThrowError("NirMalam:invalid nf being either -ve or beyond clip");

	if( ! args[5].AsBool(true) )		// user is specifying window. Check its coordinates
	{
		if ( ! ( args[7].Defined() && args[8].Defined() && args[9].Defined() && args[10].Defined() ) )
			env->ThrowError("NirMalam: lx, rx, ty, by must be defined as autoselect is not opted");

		if( args[7].AsInt() < args[1].AsInt(5) || (args[7].AsInt() & 1) != 0 )
			env->ThrowError("NirMalam: lx must not be less than xgrid and must be an even number");

		if( args[7].AsInt()  > args[8].AsInt() - 8 * args[1].AsInt(5))
			env->ThrowError("NirMalam:rx must be more than lx by atleast 8 times xgrid");

		if ( args[8].AsInt() > vi.width - 1 - args[1].AsInt(5))
			env->ThrowError("NirMalam:rx must not be more than  frame width -1 - xgrid");

		if( args[9].AsInt() < args[2].AsInt(args[1].AsInt(5)) )
			env->ThrowError("NirMalam: ty must not be less than ygrid");

		if( args[9].AsInt()  > args[10].AsInt() - 8 * args[2].AsInt(args[1].AsInt(5)))
			env->ThrowError("NirMalam:by minus ty  be atleast 8 times ygrid");

		if ( args[10].AsInt() > vi.height - 1 - args[2].AsInt(args[1].AsInt(5)))
			env->ThrowError("NirMalam:by must not be more than  frame height -1- ygrid");
	}

	else
	{
		if(args[6].AsInt((vi.height * vi.width ) /128) < (vi.height * vi.width ) /256
			|| args[6].AsInt((vi.height * vi.width ) /128) > (vi.height * vi.width ) /16)

			env->ThrowError("NirMalam:window size for auto select must be minimum %d and maximum %d",
								(vi.height * vi.width ) /256, (vi.height * vi.width ) /16);
			if(vi.height * vi.width  < 8192 )
				env->ThrowError("NirMalam:A very small frame dimension. Unsuitable for this process");


	}

	if ( args[11].AsInt(1) > 100 || args[11].AsInt(1) < 0)
		env->ThrowError("NirMalam: ry  must be between 0 and 100");

	if ( args[12].AsInt(1) > 100 || args[12].AsInt(1) < 0)
		env->ThrowError("NirMalam: gu  must be between 0 and 100");

	if ( args[13].AsInt(1) > 100 || args[13].AsInt(1) < 0)
		env->ThrowError("NirMalam: bv  must be between 0 and 100");
	
	if ( args[11].AsInt(1) == 0 && args[12].AsInt(1) == 0 && args[13].AsInt(1)  ==0)
		env->ThrowError("NirMalam: ry, gu and bv are all set to zero");

	
	return new NirMalam(args[0].AsClip(),			// clip as LeftClip
						
						args[1].AsInt(5),			// x grid points
						args[2].AsInt(args[1].AsInt(5)),			// y grid points
						
						args[3].AsBool(false),	// show variance values?
						
						args[4].AsInt(0),		// frame number from which window is to be used

						args[5].AsBool(true),	// auto select window?

						args[6].AsInt((vi.height * vi.width ) /128),	// minimum window size for auto select

						args[7].AsInt(),		// window left x;
						args[8].AsInt(),		// window right x
						args[9].AsInt(),		// window top y
						args[10].AsInt(),		// window bottom y
						
						args[11].AsInt(75),		// accent %age for R or Y
						args[12].AsInt(75),		// accent %age for G or U
						args[13].AsInt(75),		// accent %age for B or V

								env);
// Calls the constructor with the arguments provied.
}
// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.

#ifdef AVISYNTH_VERSION_2_6 
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit3(IScriptEnvironment* env, const AVS_Linkage* const vectors)
{
	// Save the server pointers.
	AVS_linkage = vectors;
#else
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
#endif
    env->AddFunction("NirMalam", "c[xgrid]i[ygrid]i[print]b[nf]i[autow]b[minw]i[lx]i[rx]i[ty]i[by]i[ry]i[gu]i[bv]i", Create_NirMalam, 0);
				
    return "NirMalam  Plugin";
}

